"""
Kernel functions and helper utilities for kernel-based calculations.

Most of the commonly-used kernel functions are already implemented in
sklearn.metrics.pairwise. The RBF kernel and pairwise Euclidean distance function is
re-implemented here as an alternative, in case sklearn is unavailable. Most, if not all
of the kernel functions defined in sklearn.metrics.pairwise should be compatible with
the functions defined here.

To use the sklearn functions, use them like so:

    >>> from sklearn.metrics import pairwise_distances
    >>> from sklearn.metrics.pairwise import pairwise_kernels
    >>> from sklearn.metrics.pairwise import rbf_kernel
    >>> from sklearn.metrics.pairwise import euclidean_distances
    >>> X = np.arange(4).reshape((2, 2))
    >>> Y = np.arange(6).reshape((3, 2))
    >>> K = gym_socks.kernel.metrics.rbf_kernel(X, Y, distance_fn=euclidean_distances)

"""

from functools import partial, reduce

import numpy as np
from numpy.linalg import inv


def euclidean_distance(
    X: np.ndarray,
    Y: np.ndarray = None,
    squared: bool = False,
) -> np.ndarray:
    """Euclidean distance function.

    The main difference between the way this function calculates euclidean
    distance over other implementations such as in sklearn.metrics.pairwise is
    that this implementation is largely agnostic to the dimensionality of the
    input data, and generally works well for high-dimensional vectors and dense
    data sets, such as state trajectories over a long time horizon.

    Args:
        X: The observations are oganized in ROWS.
        Y: The observations are oganized in ROWS.
        squared: Whether or not the result is the squared Euclidean distance.

    Returns:
        np.ndarray: Matrix of pairwise distances between points.

    """

    if Y is None:
        Y = X

    num_rows_X, num_cols_X = X.shape
    num_rows_Y, num_cols_Y = Y.shape

    def calc_dim(prev, next):
        x, y = next
        return prev + np.power(
            np.tile(y, (num_rows_X, 1)) - np.tile(x, (num_rows_Y, 1)).T,
            2,
        )

    distances = reduce(
        calc_dim, zip(X.T, Y.T), np.zeros(shape=(num_rows_X, num_rows_Y))
    )

    return distances if squared else np.sqrt(distances)


def rbf_kernel(
    X: np.ndarray,
    Y: np.ndarray = None,
    sigma: float = None,
    distance_fn=None,
) -> np.ndarray:
    """RBF kernel function.

    Computes the pairwise evaluation of the RBF kernel on each vector in X and
    Y. For instance, if X has n vecotrs, and Y has m vectors, then the result is
    an n x m matrix K where K_ij = k(xi, yj).
    ::

        X = [[--- x1 ---],
             [--- x2 ---],
             ...
             [--- xn ---]]

        Y = [[--- y1 ---],
             [--- y2 ---],
             ...
             [--- ym ---]]

    The result is a matrix,
    ::

        K = [[k(x1,y1), ..., k(x1,ym)],
             ...
             [k(xn,y1), ..., k(xn,ym)]]

    The main difference between this implementation and the rbf_kernel in
    sklearn.metrics.pairwise is that this function optionally allows you to
    specify a different distance metric in the event the data is non-Euclidean.

    Args:
        X: The observations are oganized in ROWS.
        Y: The observations are oganized in ROWS.
        sigma: Strictly positive real-valued kernel parameter.
        distance_fn: Distance function to use in the kernel evaluation.

    Returns:
        np.ndarray: Gram matrix of pairwise evaluations of the kernel function.

    """

    if distance_fn is None:
        distance_fn = euclidean_distance

    K = distance_fn(X, Y, squared=True)

    if sigma is None:
        sigma = np.median(K)
    else:
        assert sigma > 0, "sigma must be a strictly positive real value."

    K /= -2 * (sigma ** 2)
    np.exp(K, K)

    return K


def rbf_kernel_rff_transform(X, num_features=None, sigma=None):

    D = np.shape(X)[1]
    w = (1 / sigma) * np.random.standard_normal(size=(D, num_features))
    b = np.random.uniform(low=0, high=2 * np.pi, size=(num_features,))
    Z = (np.sqrt(2) / np.sqrt(num_features)) * np.cos(X @ w + b)

    return Z


def rbf_kernel_derivative(
    X: np.ndarray,
    Y: np.ndarray = None,
    sigma: float = None,
    distance_fn=None,
) -> np.ndarray:

    if distance_fn is None:
        distance_fn = euclidean_distance

    D = distance_fn(X, Y, squared=False)

    if sigma is None:
        sigma = np.median(D)
    else:
        assert sigma > 0, "sigma must be a strictly positive real valued."

    D *= 2 / (2 * (sigma ** 2))


def abel_kernel(
    X: np.ndarray,
    Y: np.ndarray = None,
    sigma: float = None,
    distance_fn=None,
) -> np.ndarray:
    """Abel kernel function.

    Args:
        X: The observations are oganized in ROWS.
        Y: The observations are oganized in ROWS.
        sigma: Strictly positive real-valued kernel parameter.
        distance_fn: Distance function to use in the kernel evaluation.

    Returns:
        np.ndarray: Gram matrix of pairwise evaluations of the kernel function.

    """

    if distance_fn is None:
        distance_fn = euclidean_distance

    K = distance_fn(X, Y, squared=False)

    if sigma is None:
        sigma = np.median(K)
    else:
        assert sigma > 0, "sigma must be a strictly positive real value."

    K /= -sigma
    np.exp(K, K)

    return K


def regularized_inverse(
    X: np.ndarray,
    Y: np.ndarray = None,
    U: np.ndarray = None,
    V: np.ndarray = None,
    regularization_param: float = None,
    kernel_fn=None,
) -> np.ndarray:
    """Regularized inverse.

    Computes the regularized matrix inverse (K + lambda * M * I)^-1.


    Args:
        X: The observations are oganized in ROWS.
        Y: The observations are oganized in ROWS.
        regularization_param: Regularization parameter, which is a strictly positive
            real value.
        kernel_fn: The kernel function is a function that returns an ndarray where
            each element is the pairwise evaluation of a kernel function. See sklearn.
            metrics.pairwise for more info. The default is the RBF kernel.

    Returns:
        np.ndarray: Regularized matrix inverse.

    """

    if Y is None:
        Y = X

    if regularization_param is None:
        regularization_param = 1 / (X.shape[0] ** 2)
    else:
        assert (
            regularization_param > 0
        ), "regularization_param must be a strictly positive real value."

    if kernel_fn is None:
        kernel_fn = partial(rbf_kernel, sigma=1)

    err_msg = "Parameters %r, %r must have the same shape." % (X, Y)
    assert X.shape == Y.shape, err_msg

    num_rows, num_cols = X.shape

    K = kernel_fn(X, Y)

    if U is not None:
        err_msg = "Parameters %r, %r must have the same sample size." % (X, U)
        assert X.shape[0] == U.shape[0], err_msg

        if V is not None:
            err_msg = "Parameters %r, %r must have the same shape." % (U, V)
            assert U.shape == V.shape, err_msg

        K = np.multiply(kernel_fn(U, V), K)

    I = np.identity(num_rows)

    return inv(K + regularization_param * num_rows * I)



__all__ = ["kernel_control_bwd", "kernel_control_fwd"]

from gym_socks.algorithms.control.kernel_control_bwd import KernelControlBwd
from gym_socks.algorithms.control.kernel_control_bwd import kernel_control_bwd

from gym_socks.algorithms.control.kernel_control_fwd import KernelControlFwd
from gym_socks.algorithms.control.kernel_control_fwd import kernel_control_fwd



from abc import ABC, abstractmethod

import numpy as np


class AlgorithmInterface(ABC):
    """Base class for algorithms.

    This class is ABSTRACT, meaning it is not meant to be instantiated directly.
    Instead, define a new class that inherits from AlgorithmInterface.

    The AlgorithmInterface is meant to mimic the sklearn estimator base class in order
    to promote a standard interface among machine learning algorithms. It requires that
    all subclasses implement a `fit` and `predict` method.

    """

    @abstractmethod
    def fit(cls, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def predict(cls, *args, **kwargs):
        raise NotImplementedError



import unittest
from unittest.mock import patch

import itertools

import gym

import gym_socks.envs

from gym_socks.envs.sample import (
    sample,
    step_sampler,
    uniform_grid_step_sampler,
    trajectory_sampler,
    sample_generator,
    transpose_sample,
)

import numpy as np


class TestSample(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.system = gym_socks.envs.NDIntegratorEnv(2)
        cls.policy = gym_socks.envs.policy.RandomizedPolicy(cls.system)

        cls.sample_space = gym.spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        cls.sample_space.seed(1)

    def test_step_sampler(cls):

        S = sample(
            sampler=step_sampler(
                system=cls.system,
                policy=cls.policy,
                sample_space=cls.sample_space,
            ),
            sample_size=10,
        )

        X, U, Y = transpose_sample(S)

        cls.assertEqual(np.array(X).shape, (10, 2))
        cls.assertEqual(np.array(U).shape, (10, 1))
        cls.assertEqual(np.array(Y).shape, (10, 2))

    def test_uniform_step_sampler(cls):

        ranges = [np.linspace(-1, 1, 5), np.linspace(-1, 1, 5)]

        S = sample(
            sampler=uniform_grid_step_sampler(
                ranges,
                system=cls.system,
                policy=cls.policy,
                sample_space=cls.sample_space,
            ),
            sample_size=25,
        )

        X, U, Y = transpose_sample(S)

        cls.assertEqual(np.array(X).shape, (25, 2))
        cls.assertEqual(np.array(U).shape, (25, 1))
        cls.assertEqual(np.array(Y).shape, (25, 2))

        groundTruth = list(itertools.product(*ranges))
        cls.assertTrue(np.all(np.equal(np.array(X), groundTruth)))

        # Sample more than ranges.
        S = sample(
            sampler=uniform_grid_step_sampler(
                ranges,
                system=cls.system,
                policy=cls.policy,
                sample_space=cls.sample_space,
            ),
            sample_size=50,
        )

        X, U, Y = transpose_sample(S)

        cls.assertEqual(np.array(X).shape, (50, 2))
        cls.assertEqual(np.array(U).shape, (50, 1))
        cls.assertEqual(np.array(Y).shape, (50, 2))

        cls.assertTrue(np.all(np.equal(np.array(X[:25]), np.array(X[25:]))))

    def test_trajectory_sampler(cls):

        S = sample(
            sampler=trajectory_sampler(
                system=cls.system,
                policy=cls.policy,
                sample_space=cls.sample_space,
            ),
            sample_size=10,
        )

        cls.assertEqual(np.array(S[0][0]).shape, (2,))
        cls.assertEqual(np.array(S[0][1]).shape, (16, 1))
        cls.assertEqual(np.array(S[0][2]).shape, (16, 2))

        X, U, Y = transpose_sample(S)

        cls.assertEqual(np.array(X).shape, (10, 2))
        cls.assertEqual(np.array(U).shape, (10, 16, 1))
        cls.assertEqual(np.array(Y).shape, (10, 16, 2))

        ST = gym_socks.envs.sample.reshape_trajectory_sample(S)
        X, U, Y = transpose_sample(ST)

        cls.assertEqual(np.array(X).shape, (10, 2))
        cls.assertEqual(np.array(U).shape, (10, 16))
        cls.assertEqual(np.array(Y).shape, (10, 32))

    def test_custom_sampler_as_function(cls):
        @sample_generator
        def custom_sampler():
            state = cls.sample_space.sample()
            action = cls.system.action_space.sample()

            cls.system.state = state
            next_state, cost, done, _ = cls.system.step(action)

            return (state, action, next_state)

        #   ^^^^^^

        S = sample(
            sampler=custom_sampler,
            sample_size=10,
        )

        X, U, Y = transpose_sample(S)

        cls.assertEqual(np.array(X).shape, (10, 2))
        cls.assertEqual(np.array(U).shape, (10, 1))
        cls.assertEqual(np.array(Y).shape, (10, 2))

    def test_custom_sampler_as_generator(cls):
        @sample_generator
        def custom_sampler():
            state = cls.sample_space.sample()
            action = cls.system.action_space.sample()

            cls.system.state = state
            next_state, cost, done, _ = cls.system.step(action)

            yield (state, action, next_state)

        #   ^^^^^

        S = sample(
            sampler=custom_sampler,
            sample_size=10,
        )

        X, U, Y = transpose_sample(S)

        cls.assertEqual(np.array(X).shape, (10, 2))
        cls.assertEqual(np.array(U).shape, (10, 1))
        cls.assertEqual(np.array(Y).shape, (10, 2))



from sacred import Ingredient

import gym
from sacred.utils import convert_camel_case_to_snake_case

from gym_socks.envs.dynamical_system import DynamicalSystem

from gym_socks.envs.sample import sample
from gym_socks.envs.sample import sample_generator
from gym_socks.envs.sample import step_sampler
from gym_socks.envs.sample import uniform_grid
from gym_socks.envs.sample import uniform_grid_step_sampler

import numpy as np

from examples.ingredients.common import parse_array, box_factory
from examples.ingredients.common import grid_ranges
from examples.ingredients.common import grid_sample_size

forward_reach_ingredient = Ingredient("forward_reach")


@forward_reach_ingredient.config
def _config():
    """Forward reachability configuration variables."""

    test_points = {
        "lower_bound": -1,
        "upper_bound": 1,
        "grid_resolution": 50,
    }


@forward_reach_ingredient.capture
def compute_test_point_ranges(shape, test_points, _log):

    lower_bound = test_points["lower_bound"]
    upper_bound = test_points["upper_bound"]

    grid_resolution = test_points["grid_resolution"]

    test_point_space = box_factory(
        lower_bound,
        upper_bound,
        shape=shape,
        dtype=np.float32,
    )

    xi = grid_ranges(test_point_space, grid_resolution)

    return xi


@forward_reach_ingredient.capture
def generate_test_points(shape, _log):
    """Generate test points."""

    _log.info("Generating test points.")

    xi = compute_test_point_ranges(shape)
    T = uniform_grid(xi)

    return T



from sacred import Ingredient

import gym

from gym_socks.envs.dynamical_system import DynamicalSystem

from gym_socks.envs.sample import sample
from gym_socks.envs.sample import sample_generator
from gym_socks.envs.sample import step_sampler
from gym_socks.envs.sample import uniform_grid
from gym_socks.envs.sample import uniform_grid_step_sampler

import numpy as np

from examples.ingredients.common import parse_array, box_factory
from examples.ingredients.common import grid_ranges
from examples.ingredients.common import grid_sample_size

backward_reach_ingredient = Ingredient("backward_reach")


@backward_reach_ingredient.config
def _config():
    """Backward reachability configuration variables."""

    problem = "THT"

    constraint_tube_bounds = {"lower_bound": -1, "upper_bound": 1}
    target_tube_bounds = {"lower_bound": -0.5, "upper_bound": 0.5}

    test_points = {
        "lower_bound": -1,
        "upper_bound": 1,
        "grid_resolution": 25,
    }


def generate_tube(env: DynamicalSystem, bounds: dict):
    """Generate a stochastic reachability tube using config.

    This function computes a stochastic reachability tube using the tube configuration.

    Args:
        env: The dynamical system model.
        bounds: The bounds of the tube. Specified as a dictionary.

    Returns:
        A list of spaces indexed by time.

    """

    tube_lb = bounds["lower_bound"]
    tube_ub = bounds["upper_bound"]

    tube_lb = parse_array(tube_lb, shape=(env.num_time_steps,), dtype=np.float32)
    tube_ub = parse_array(tube_ub, shape=(env.num_time_steps,), dtype=np.float32)

    tube = []
    for i in range(env.num_time_steps):
        tube_t = box_factory(
            tube_lb[i],
            tube_ub[i],
            shape=env.state_space.shape,
            dtype=np.float32,
        )
        tube.append(tube_t)

    return tube


@backward_reach_ingredient.capture
def compute_test_point_ranges(env, test_points):

    lower_bound = test_points["lower_bound"]
    upper_bound = test_points["upper_bound"]

    grid_resolution = test_points["grid_resolution"]

    test_point_space = box_factory(
        lower_bound,
        upper_bound,
        shape=env.state_space.shape,
        dtype=np.float32,
    )

    xi = grid_ranges(test_point_space, grid_resolution)

    return xi


@backward_reach_ingredient.capture
def generate_test_points(env, _log):
    """Generate test points to evaluate the safety probabilities."""
    _log.debug("Generating test points.")
    xi = compute_test_point_ranges(env)
    T = uniform_grid(xi)

    return T



"""Nonholonomic vehicle system."""

import gym
from gym_socks.envs.dynamical_system import DynamicalSystem

import numpy as np

from scipy.integrate import solve_ivp


class NonholonomicVehicleEnv(DynamicalSystem):
    """Nonholonomic vehicle system.

    A nonholonomic vehicle (car-like) is typically modeled using what are known as
    "unicycle" dynamics. It is useful for modeling vehicles which can move forward and
    backward, and incorporates a steering angle or heading. The inputs are the velocity
    and change in steering angle.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            observation_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(3,), dtype=np.float32
            ),
            state_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(3,), dtype=np.float32
            ),
            action_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(2,), dtype=np.float32
            ),
            *args,
            **kwargs
        )

    def step(self, action, time=0):
        err_msg = "%r (%s) invalid" % (action, type(action))
        assert self.action_space.contains(action), err_msg

        disturbance = self.generate_disturbance(time, self.state, action)

        # solve the initial value problem
        if self._euler is True:
            next_state = self.state + self.sampling_time * self.dynamics(
                time, self.state, action, disturbance
            )
            self.state = next_state
        else:
            # solve the initial value problem
            sol = solve_ivp(
                self.dynamics,
                [time, time + self.sampling_time],
                self.state,
                args=(
                    action,
                    disturbance,
                ),
            )
            *_, self.state = sol.y.T

        # correct the angle
        if np.abs(self.state[2]) >= 2 * np.pi:
            self.state[2] %= 2 * np.pi

        observation = self.generate_observation(time, self.state, action)

        cost = self.cost(time, self.state, action)

        done = False
        info = {}

        return observation, cost, done, info

    def generate_disturbance(self, time, state, action):
        w = self.np_random.standard_normal(size=self.state_space.shape)
        return 1e-2 * np.array(w)

    def dynamics(self, time, state, action, disturbance):

        x1, x2, x3 = state
        u1, u2 = action
        w1, w2, w3 = disturbance

        dx1 = u1 * np.sin(x3) + w1
        dx2 = u1 * np.cos(x3) + w2
        dx3 = u2 + w3

        return np.array([dx1, dx2, dx3], dtype=np.float32)



"""Computation timer."""

import logging
from time import perf_counter

logger = logging.getLogger("timer")


class ComputationTimer(object):
    """Computation timer.

    Simple timer class for measuring the time of an algorithm and displaying the result.

    Example:

        >>> from examples._computation_timer import ComputationTimer
        >>> with ComputationTimer():
        ...     # run algorithm
        computation time: 3.14159 s

    """

    def __init__(self) -> None:
        self._start_time = None

    def __enter__(self):
        logger.debug("Starting timer.")
        self._start_time = perf_counter()
        return self

    def __exit__(self, *exc_info):
        self.log_time()
        self._start_time = None

    def log_time(self):
        """Output the computation time to the log."""
        elapsed_time = perf_counter() - self._start_time
        logger.debug("Stopping timer.")
        logger.info(f"computation time: {elapsed_time} s")



"""Separating kernel classifier.

Separating kernel classifier, useful for forward stochastic reachability analysis.

References:
    .. [1] `Learning sets with separating kernels, 2014
            De Vito, Ernesto, Lorenzo Rosasco, and Alessandro Toigo.
            Applied and Computational Harmonic Analysis 37(2)`_

    .. [2] `Learning Approximate Forward Reachable Sets Using Separating Kernels, 2021
            Adam J. Thorpe, Kendric R. Ortiz, Meeko M. K. Oishi
            Learning for Dynamics and Control,
            <https://arxiv.org/abs/2011.09678>`_

"""

from functools import partial

from gym_socks.algorithms.algorithm import AlgorithmInterface
from gym_socks.kernel.metrics import abel_kernel, regularized_inverse

import numpy as np


class SeparatingKernelClassifier(AlgorithmInterface):
    """Separating kernel classifier.

    A kernel-based support classifier for unknown distributions. Given a set of data taken iid from the distribution, the `SeparatingKernelClassifier` constructs a kernel-based classifier of the support of the distribution based on the theory of separating kernels.

    Note:
        The sample used by the classifier is from the marginal distribution, not the
        joint or conditional. Thus, the data should be an array of points organized such
        that each point occupies a single row in a 2D-array.

    Args:
        kernel_fn: The kernel function used by the classifier.
        regularization_param: The regularization parameter used in the regularized
            least-squares problem. Determines the smoothness of the solution.

    Example:
        >>> from gym_socks.algorithms.reach import SeparatingKernelClassifier
        >>> from gym_socks.kernel.metrics import abel_kernel
        >>> from functools import partial
        >>> kernel_fn = partial(abel_kernel, sigma=0.1)
        >>> classifier = SeparatingKernelClassifier(kernel_fn)
        >>> classifier.fit(S)
        >>> classifications = classifier.predict(T)

    """

    def __init__(
        self,
        kernel_fn=None,
        regularization_param=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.kernel_fn = kernel_fn
        self.regularization_param = regularization_param

    def _validate_params(self, S):

        if self.kernel_fn is None:
            self.kernel_fn = partial(abel_kernel, sigma=0.1)

        if self.regularization_param is None:
            self.regularization_param = 1

    def _validate_data(self, X):

        if X is None:
            raise ValueError("Must supply a sample.")

    def fit(self, X: np.ndarray):
        """Fit separating kernel classifier.

        Args:
            X: Data drawn from distribution.

        Returns:
            self: Instance of SeparatingKernelClassifier

        """

        self._validate_data(X)
        X = np.array(X)

        self.X = X

        K = self.kernel_fn(X)

        self.W = regularized_inverse(
            X, kernel_fn=self.kernel_fn, regularization_param=self.regularization_param
        )

        self.tau = 1 - np.min(np.diagonal((1 / len(X)) * K.T @ self.W @ K))

        return self

    def predict(self, T: np.ndarray) -> list:
        """Predict using the separating kernel classifier.

        Args:
            T: Evaluation points where the separating kernel classifier is evaluated.

        Returns:
            Boolean indicator of classifier.

        """

        self._validate_data(T)
        T = np.array(T)

        K = self.kernel_fn(self.X, T)

        C = np.diagonal((1 / len(self.X)) * K.T @ self.W @ K)

        return C >= 1 - self.tau



__all__ = ["logging"]

from typing import Generator
import gym

import numpy as np


def normalize(v: np.ndarray) -> np.ndarray:
    """Normalize.

    Small utility function for normalizing a matrix or a vector. Divides the matrix
    (vector) by the sum of the matrix rows (vector). Used primarily by the stochastic
    reachability algorithms.

    Args:
        v: Matrix or vector to normalize.

    Returns:
        Normalized matrix or vector.

    """
    return v / np.sum(v, axis=0)


def _gym_space_contains(points: np.ndarray, set: gym.spaces.Box) -> np.ndarray:
    """Lightweight indicator for gym.spaces.Box sets.

    Provides a lightweight proxy for the `contains` function in `gym.spaces.Box`. The
    original function does type checking, casting, and shape checking, which is slow.

    Args:
        points: Points to evaluate the indicator at.
        set: Set over which the indicator function is defined. Should be either a
            `gym.spaces.Box` or a function which returns 0 or 1 depending on whether the
            points are inside or outside the set.

    Returns:
        Boolean labels of whether the points are inside or outside the set.

    """

    _l = set.low
    _h = set.high

    return np.array(
        np.all(points >= _l, axis=1) & np.all(points <= _h, axis=1), dtype=bool
    )


def indicator_fn(points: np.ndarray, set: any) -> np.ndarray:
    """Indicator function for generic sets.

    If the set is not an instance of `gym.spaces.Box`, then `set` should be a callable
    function. Any easy way to enable this is to create a wrapper. The function should
    return a 0 or 1 if the points are outside or inside the set, respectively.

    Args:
        points: Points to evaluate the indicator at.
        set: Set over which the indicator function is defined. Should be either a
            `gym.spaces.Box` or a function which returns 0 or 1 depending on whether the
            points are inside or outside the set.

    Returns:
        Boolean labels of whether the points are inside or outside the set.

    """

    if isinstance(set, gym.spaces.Box):
        return _gym_space_contains(points, set)

    return np.array(set(points), dtype=bool)


def generate_batches(num_elements, batch_size) -> Generator[slice, None, None]:
    """Generate batches.

    Batch generation function to split a list into smaller batches (slices). Generates
    `slice` objects, which can be iterated over in a for loop.

    Args:
        num_elements: Length of the list.
        batch_size: Maximum size of the batches. If the last batch is smaller than
            `batch_size`, then the final batch will be the length of the remaining
            elements.

    Yields:
        A slice object.

    """

    start = 0

    for _ in range(num_elements // batch_size):
        end = start + batch_size

        yield slice(start, end)

        start = end

    if start < num_elements:
        yield slice(start, num_elements)



"""Plotting ingredient.

This ingredient is used for experiments with a plotting component.

"""

import os
import logging

import numpy as np

from sacred import Ingredient

plotting_ingredient = Ingredient("plot_cfg")


@plotting_ingredient.config
def _config():
    rc_params = dict()
    rc_params_filename = os.path.abspath("examples/ingredients/matplotlibrc")

    plot_filename = "results/plot.png"


@plotting_ingredient.capture
def update_rc_params(matplotlib, rc_params, rc_params_filename):
    custom_rc_params = matplotlib.rc_params_from_file(
        fname=rc_params_filename, use_default_template=True
    )
    matplotlib.rcParams.update(custom_rc_params)

    for key, value in rc_params.items():
        matplotlib.rc(key, **value)



"""ND Integrator system."""

import gym
from gym_socks.envs.dynamical_system import DynamicalSystem

import numpy as np


class NDIntegratorEnv(DynamicalSystem):
    """ND integrator system.

    An integrator system is an extremely simple dynamical system model, typically used
    to model a single variable and its higher order derivatives, where the input is
    applied to the highest derivative term, and is "integrated" upwards.

    A 2D Integrator system, for example, corresponds to the position and velocity
    components of a variable, where the input is applied to the velocity and then
    integrates upward to the position variable. Chaining two 2D integrator systems can
    model a system with x/y position and velocity.

    Args:
        dim: The dimension of the integrator system.

    Example:

        >>> from gym_socks.envs import NDIntegratorEnv
        >>> env = NDIntegratorEnv(dim=2)
        >>> env.reset()
        >>> num_steps = env.num_time_steps
        >>> for i in range(num_steps):
        ...     action = env.action_space.sample()
        ...     obs, reward, done, _ = env.step(action)
        >>> env.close()

    """

    def __init__(self, dim: int = 1, *args, **kwargs):
        super().__init__(
            observation_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(dim,), dtype=np.float32
            ),
            state_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(dim,), dtype=np.float32
            ),
            action_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
            ),
            *args,
            **kwargs
        )

        self.time_horizon = 4
        self.sampling_time = 0.25

    def generate_disturbance(self, time, state, action):
        w = self.np_random.standard_normal(size=self.state_space.shape)
        return 1e-2 * np.array(w)

    def dynamics(self, time, state, action, disturbance):
        _, *x = state
        return np.array([*x, *action], dtype=np.float32) + disturbance



import gym

import numpy as np

from sacred import Ingredient

from gym_socks.envs.dynamical_system import DynamicalSystem

cwh_ingredient = Ingredient("cwh")


@cwh_ingredient.config
def _config():

    norm_order = 2
    squared = False

    target_state = [0, 0, 0, 0]

    terminal_constraint = {
        "lower_bound": [-0.2, -0.2, -0.05, -0.05],
        "upper_bound": [-0.2, 0, 0.05, 0.05],
    }


@cwh_ingredient.capture
def make_cost(env: DynamicalSystem, target_state, norm_order, squared):
    def _cost_fn(time: int = 0, state: np.ndarray = None) -> float:
        """CWH cost function.

        The cost is defined such that we seek to minimize the distance from the system
        to the origin. This would indicate a fully "docked" spacecraft with zero
        terminal velocity.

        """

        dist = state - np.array([target_state], dtype=np.float32)
        result = np.linalg.norm(dist, ord=norm_order, axis=1)
        if squared is True:
            result = np.power(result, 2)
        return result

    return _cost_fn


@cwh_ingredient.capture
def make_constraint(env: DynamicalSystem):
    def _constraint_fn(time: int = 0, state: np.ndarray = None) -> float:
        """CWH constraint function.

        The CWH constraint function is defined as the line of sight (LOS) cone from the
        spacecraft, where the velocity components are sufficiently small.

        Note:
            The constraints are written in terms of indicator functions, which return a
            one if the sample is in the LOS cone (or in the target space) and a zero if
            it is not, but the optimal control problem is defined such that the
            constraints are satisfied if the function is less than or equal to zero.
            Thus, we use the following algebraic manipulation::

                1[A](x) >= 1
                - 1[A](x) <= -1
                - 1[A](x) + 1 <= 0

        """

        # Terminal constraint.
        if time < env.num_time_steps - 1:
            satisfies_constraints = (
                -np.array(
                    (np.abs(state[:, 0]) < np.abs(state[:, 1]))
                    & (np.abs(state[:, 2]) <= 0.05)
                    & (np.abs(state[:, 3]) <= 0.05),
                    dtype=np.float32,
                )
                + 1
            )

            return np.round(satisfies_constraints, decimals=2)

        # LOS constraint.
        else:
            satisfies_constraints = (
                -np.array(
                    (np.abs(state[:, 0]) < 0.2)
                    & (state[:, 1] >= -0.2)
                    & (state[:, 1] <= 0)
                    & (np.abs(state[:, 2]) <= 0.05)
                    & (np.abs(state[:, 3]) <= 0.05),
                    dtype=np.float32,
                )
                + 1
            )

            return np.round(satisfies_constraints, decimals=2)

    return _constraint_fn



"""Sampling methods.

This file contains a collection of sampling methods. The core principle is to define a
function that returns a single observation (either via return or yield) from a
probability measure. Then, the `sample_generator` is a decorator, which converts a
function that returns a single observation into a generator, that can be sampled using
`islice`.

Example:
    Sample the stochastic kernel of a dynamical system (i.e. the state transition
    probability kernel).

        >>> env = NdIntegrator(2)
        >>> sample_space = gym.spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        >>> sampler = step_sampler(
        ...     system=env, policy=RandomizedPolicy(env), sample_space=sample_space
        ... )
        >>> S = sample(sampler=sampler, sample_size=100)

The main reason for this setup is to allow for 'observation functions' which have
different structures, e.g.
* a function which `return`s an observation,
* an infinite generator which `yield`s observations, and
* a finite generator that `yield`s observations.

"""

from inspect import isgeneratorfunction
from functools import partial, wraps
from itertools import islice, product

import gym

from gym_socks.envs.dynamical_system import DynamicalSystem
from gym_socks.envs.policy import BasePolicy
from gym_socks.envs.policy import RandomizedPolicy
from gym_socks.envs.policy import ConstantPolicy

import numpy as np
from scipy.integrate import solve_ivp


def sample_generator(fun):
    """Sample generator decorator.

    Converts a sample function into a generator function. Any function that returns a
    single observation (as a tuple) can be converted into a sample generator.

    Args:
        fun: Sample function that returns or yields an observation.

    Returns:
        A function that can be used to `islice` a sample from the sample generator.

    Example:

        >>> from itertools import islice
        >>> from gym_socks.envs.sample import sample_generator
        >>> @sample_generator
        ... def custom_sampler(system, policy, sample_space):
        ...     system.state = sample_space.sample()
        ...     action = policy(state=state)
        ...     next_state, *_ = system.step(action)
        ...     yield (system.state, action, next_state)
        >>> S = list(islice(custom_sampler(), 100))

    """

    @wraps(fun)
    def _wrapper(*args, **kwargs):
        while True:
            if isgeneratorfunction(fun):
                yield from fun(*args, **kwargs)
            else:
                yield fun(*args, **kwargs)

    return _wrapper


def step_sampler(
    system: DynamicalSystem = None,
    policy: BasePolicy = None,
    sample_space: gym.Space = None,
):
    """Default sampler (one step).

    Args:
        system: The system to sample from.
        policy: The policy applied to the system during sampling.
        sample_space: The space where initial conditions are drawn from.

    Returns:
        A generator function that yields system observations as tuples.

    """

    @sample_generator
    def _sample_generator():
        state = sample_space.sample()
        action = policy(state=state)

        system.state = state
        next_state, cost, done, _ = system.step(action)

        yield (state, action, next_state)

    return _sample_generator


def uniform_grid(xi):
    """Create a uniform grid from a list of ranges.

    Args:
        xi: List of ranges.

    Returns:
        Grid of points (the product of all points in ranges).

    Example:

        >>> import numpy as np
        >>> from gym_socks.envs.sample import uniform_grid
        >>> grid = uniform_grid([np.linspace(-1, 1, 5), np.linspace(-1, 1, 5)])

    """

    return list(product(*xi))


def uniform_grid_step_sampler(
    xi: list,
    system: DynamicalSystem = None,
    policy: BasePolicy = None,
    sample_space: gym.Space = None,
):
    """Uniform sampler (one step).

    Args:
        xi: List of ranges.
        system: The system to sample from.
        policy: The policy applied to the system during sampling.
        sample_space: The space where initial conditions are drawn from.

    Returns:
        A generator function that yields system observations as tuples.

    """

    xc = uniform_grid(xi)

    @sample_generator
    def _sample_generator():
        for point in xc:
            state = point
            action = policy(state=state)

            system.state = state
            next_state, cost, done, _ = system.step(action)

            yield (state, action, next_state)

    return _sample_generator


def sequential_action_sampler(
    ui: list,
    env: DynamicalSystem,
    sample_space: gym.Space,
    sampler,
):

    uc = uniform_grid(ui)

    @sample_generator
    def _wrapper(*args, **kwargs):
        for action in uc:
            _policy = ConstantPolicy(env, constant=action)

            sampler_instance = sampler(
                system=env,
                policy=_policy,
                sample_space=sample_space,
            )

            yield sampler_instance(*args, **kwargs)

    return _wrapper


def trajectory_sampler(
    system: DynamicalSystem = None,
    policy: BasePolicy = None,
    sample_space: gym.Space = None,
):
    """Default trajectory sampler.

    Args:
        system: The system to sample from.
        policy: The policy applied to the system during sampling.
        sample_space: The space where initial conditions are drawn from.

    Returns:
        A generator function that yields system observations as tuples.

    """

    @sample_generator
    def _sample_generator():
        state = sample_space.sample()

        state_sequence = []
        action_sequence = []

        system.state = state

        time = 0
        for t in range(system.num_time_steps):
            action = policy(time=time, state=system.state)
            next_state, cost, done, _ = system.step(action)

            state_sequence.append(next_state)
            action_sequence.append(action)

            time += 1

        yield (state, action_sequence, state_sequence)

    return _sample_generator


def sample(sampler=None, sample_size: int = None, *args, **kwargs):
    """Generate a sample using the sample generator.

    Args:
        sampler: Sample generator function.
        sample_size: Size of the sample.

    Returns:
        list of tuples

    """

    if sampler is None:
        raise ValueError("Must supply a sample function.")

    if sample_size is None:
        raise ValueError("Must supply a sample size.")

    if not isinstance(sample_size, int) or sample_size < 1:
        raise ValueError("Sample size must be a positive integer.")

    return list(islice(sampler(*args, **kwargs), sample_size))


def transpose_sample(sample):
    """Transpose the sample.

    By default, a sample should be a list of tuples of the form::

        S = [(x_1, y_1), ..., (x_n, y_n)]

    For most algorithms, we need to isolate the sample components (e.g. all x's).
    This function converts a sample from a list of tuples to a tuple of lists::

        S_T = ([x_1, ..., x_n], [y_1, ..., y_n])

    This can then be unpacked as: ``X, Y = S_T``

    Args:
        sample: list of tuples

    Returns:
        tuple of lists

    """

    return tuple(map(list, zip(*sample)))


def reshape_trajectory_sample(sample):
    """Reshapes trajectory samples.

    Often, trajectory samples are organized such that the "trajectory" components are a
    2D array of points indexed by time. However, for kernel methods, we typically
    require that the trajectories be concatenated into a single vector (1D array)::

        [[x1], [x2], ..., [xn]] -> [x1, x2, ..., xn]

    This function converts the sample so that the trajectories are 1D arrays.

    Args:
        sample: list of tuples

    Returns:
        List of tuples, where the components of the tuples are flattened.

    """
    sample_size = len(sample)
    _S = transpose_sample(sample)

    return zip(*[np.reshape(item, (sample_size, -1)) for item in _S])



"""Forward in time stochastic optimal control.

References:
    .. [1] `Stochastic Optimal Control via
            Hilbert Space Embeddings of Distributions, 2021
            Adam J. Thorpe, Meeko M. K. Oishi
            IEEE Conference on Decision and Control,
            <https://arxiv.org/abs/2103.12759>`_
"""

from functools import partial

import gym_socks

from gym_socks.envs.policy import BasePolicy
from gym_socks.algorithms.control.control_common import compute_solution
from gym_socks.kernel.metrics import rbf_kernel, regularized_inverse
from gym_socks.utils.logging import ms_tqdm, _progress_fmt

import numpy as np

from tqdm.auto import tqdm


def kernel_control_fwd(
    S: np.ndarray,
    A: np.ndarray,
    cost_fn=None,
    constraint_fn=None,
    heuristic: bool = False,
    regularization_param: float = None,
    kernel_fn=None,
    verbose: bool = True,
):
    """Stochastic optimal control policy forward in time.

    Computes the optimal control action at each time step in a greedy fashion. In other
    words, at each time step, the policy optimizes the cost function from the current
    state. It does not "look ahead" in time.

    Args:
        S: Sample taken iid from the system evolution.
        A: Collection of admissible control actions.
        cost_fn: The cost function. Should return a real value.
        constraint_fn: The constraint function. Should return a real value.
        heuristic: Whether to use the heuristic solution instead of solving the LP.
        regularization_param: Regularization prameter for the regularized least-squares
            problem used to construct the approximation.
        kernel_fn: The kernel function used by the algorithm.
        verbose: Whether the algorithm should print verbose output.

    Returns:
        The policy.
    """

    alg = KernelControlFwd(
        cost_fn=cost_fn,
        constraint_fn=constraint_fn,
        heuristic=heuristic,
        regularization_param=regularization_param,
        kernel_fn=kernel_fn,
        verbose=verbose,
    )
    alg.train(S, A)

    return alg


class KernelControlFwd(BasePolicy):
    """Stochastic optimal control policy forward in time.

    Computes the optimal control action at each time step in a greedy fashion. In other
    words, at each time step, the policy optimizes the cost function from the current
    state. It does not "look ahead" in time.

    Args:
        cost_fn: The cost function. Should return a real value.
        constraint_fn: The constraint function. Should return a real value.
        heuristic: Whether to use the heuristic solution instead of solving the LP.
        regularization_param: Regularization prameter for the regularized least-squares
            problem used to construct the approximation.
        kernel_fn: The kernel function used by the algorithm.
        verbose: Whether the algorithm should print verbose output.

    """

    def __init__(
        self,
        cost_fn=None,
        constraint_fn=None,
        heuristic: bool = False,
        regularization_param: float = None,
        kernel_fn=None,
        verbose: bool = True,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.cost_fn = cost_fn
        self.constraint_fn = constraint_fn

        self.heuristic = heuristic

        self.kernel_fn = kernel_fn
        self.regularization_param = regularization_param

        self.verbose = verbose

    def _validate_params(self, S=None, A=None):

        if S is None:
            raise ValueError("Must supply a sample.")

        if A is None:
            raise ValueError("Must supply a sample.")

        if self.kernel_fn is None:
            self.kernel_fn = partial(rbf_kernel, sigma=0.1)

        if self.regularization_param is None:
            self.regularization_param = 1

        if self.cost_fn is None:
            raise ValueError("Must supply a cost function.")

        if self.constraint_fn is None:
            # raise ValueError("Must supply a constraint function.")
            self._constrained = False
        else:
            self._constrained = True

    def _validate_data(self, S):

        if S is None:
            raise ValueError("Must supply a sample.")

    def train(self, S: np.ndarray, A: np.ndarray):
        """Train the algorithm.

        Args:
            S: Sample taken iid from the system evolution.
            A: Collection of admissible control actions.

        Returns:
            self: An instance of the KernelControlFwd algorithm class.

        """

        self._validate_data(S)
        self._validate_data(A)
        self._validate_params(S=S, A=A)

        X, U, Y = gym_socks.envs.sample.transpose_sample(S)
        X = np.array(X)
        U = np.array(U)
        Y = np.array(Y)

        self.X = X

        A = np.array(A)

        self.A = A

        gym_socks.logger.debug("Computing matrix inverse.")
        self.W = regularized_inverse(
            X,
            U=U,
            kernel_fn=self.kernel_fn,
            regularization_param=self.regularization_param,
        )

        gym_socks.logger.debug("Computing covariance matrix.")
        self.CUA = self.kernel_fn(U, A)

        gym_socks.logger.debug("Computing cost function.")
        self.cost_fn = partial(self.cost_fn, state=Y)

        gym_socks.logger.debug("Computing constraint function.")
        if self.constraint_fn is not None:
            self.constraint_fn = partial(self.constraint_fn, state=Y)

        return self

    def __call__(self, time=0, state=None, *args, **kwargs):

        if state is None:
            print("Must supply a state to the policy.")
            return None

        T = np.array(state)

        # Compute covariance matrix.
        CXT = self.kernel_fn(self.X, T)
        # Compute beta.
        betaXT = self.W @ (CXT * self.CUA)

        # Compute cost vector.
        C = np.matmul(np.array(self.cost_fn(time=time), dtype=np.float32), betaXT)

        # Compute constraint vector.
        D = None
        if self.constraint_fn is not None:
            D = np.matmul(
                np.array(self.constraint_fn(time=time), dtype=np.float32), betaXT
            )

        # Compute the solution to the LP.
        gym_socks.logger.debug("Computing solution to the LP.")
        sol = compute_solution(C, D, heuristic=self.heuristic)
        idx = np.argmax(sol)
        return np.array(self.A[idx], dtype=np.float32)



from sacred import Ingredient

import gym

import numpy as np

tracking_ingredient = Ingredient("tracking")


@tracking_ingredient.config
def _config():

    path_amplitude = 0.5
    path_period = 2.0

    norm_order = 2
    squared = True


@tracking_ingredient.capture
def compute_target_trajectory(num_steps, path_amplitude, path_period):
    """Computes the target trajectory to follow.

    The default trajectory is a V-shaped path based on a triangle function. The amplitude and period are set by the config.

    Args:
        amplitude : The amplitude of the triangle function.
        period : The period of the triangle function.
        sampling_time : The sampling time of the dynamical system. Ensures that there
            are a number of points in the target trajectory equal to the number of time
            steps in the simulation.
        time_horizon : The time horizon of the dynamical system. Ensures that there are
            a number of points in the target trajectory equal to the number of time
            steps in the simulation.

    Returns:
        target_trajectory : The target trajectory as a list of points.

    """

    a = path_amplitude
    p = path_period

    target_trajectory = [
        [
            (x * 0.1) - 1.0,
            4 * a / p * np.abs((((((x * 0.1) - 1.0) - p / 2) % p) + p) % p - p / 2) - a,
        ]
        for x in range(num_steps)
    ]

    return target_trajectory


@tracking_ingredient.capture
def make_cost(target_trajectory, norm_order, squared):
    def _tracking_cost(time: int = 0, state: np.ndarray = None) -> float:
        """Tracking cost function.

        The goal is to minimize the distance of the x/y position of the vehicle to the
        'state' of the target trajectory at each time step.

        Args:
            time : Time of the simulation. Used for time-dependent cost functions.
            state : State of the system.

        Returns:
            cost : Real-valued cost.

        """

        dist = state[:, :2] - np.array([target_trajectory[time]])
        result = np.linalg.norm(dist, ord=norm_order, axis=1)
        if squared is True:
            result = np.power(result, 2)
        return result

    return _tracking_cost



from typing import ForwardRef
import unittest
from unittest import mock
from unittest.mock import patch

import gym

from gym_socks.envs import NDIntegratorEnv

import numpy as np

from examples.ingredients.forward_reach_ingredient import forward_reach_ingredient
from examples.ingredients.forward_reach_ingredient import compute_test_point_ranges


class TestForwardReachIngredient(unittest.TestCase):
    def test_forward_reach_config(cls):
        """Test the ingredient config."""
        ingredient = forward_reach_ingredient
        cls.assertEqual(len(ingredient.configurations), 1)


class TestComputeTestPointRanges(unittest.TestCase):
    def test_compute_test_point_ranges(cls):
        """Test that compute_test_point_ranges gives correct ranges."""

        test_points = {
            "lower_bound": -1,
            "upper_bound": 1,
            "grid_resolution": 25,
        }

        result = compute_test_point_ranges(shape=(2,), test_points=test_points)

        groundTruth = [np.linspace(-1, 1, 25), np.linspace(-1, 1, 25)]

        cls.assertTrue(np.array_equal(result, groundTruth))



import unittest

from functools import partial

import gym

import gym_socks.kernel.metrics

import numpy as np

from sklearn.metrics.pairwise import linear_kernel
from sklearn.metrics.pairwise import polynomial_kernel
from sklearn.metrics.pairwise import rbf_kernel
from sklearn.metrics.pairwise import laplacian_kernel

from sklearn.metrics.pairwise import euclidean_distances

sklearn_kernel_list = [linear_kernel, polynomial_kernel, rbf_kernel, laplacian_kernel]


class TestEuclideanDistance(unittest.TestCase):
    def test_euclidean_distance(cls):
        """Test that euclidean distance computes correctly."""

        X = np.arange(4).reshape((2, 2))
        Y = np.arange(4).reshape((2, 2))

        groundTruth = np.array([[0.0, 8.0], [8.0, 0.0]], dtype=np.float32)

        distance = gym_socks.kernel.metrics.euclidean_distance(Y, squared=True)
        cls.assertTrue(np.allclose(distance, groundTruth))

        distance = gym_socks.kernel.metrics.euclidean_distance(X, Y, squared=True)
        cls.assertTrue(np.allclose(distance, groundTruth))

        groundTruth = euclidean_distances(X, Y, squared=True)

        distance = gym_socks.kernel.metrics.euclidean_distance(Y, squared=True)
        cls.assertTrue(np.allclose(distance, groundTruth))

        distance = gym_socks.kernel.metrics.euclidean_distance(X, Y, squared=True)
        cls.assertTrue(np.allclose(distance, groundTruth))


class TestRBFKernel(unittest.TestCase):
    def test_rbf_kernel_same_size(cls):
        """Test that RBF kernel computes correctly with same size inputs."""

        Y = np.arange(4).reshape((2, 2))

        groundTruth = np.array(
            [[1.0, 0.77880078], [0.77880078, 1.0]],
            dtype=np.float32,
        )

        K = gym_socks.kernel.metrics.rbf_kernel(Y)
        cls.assertTrue(np.allclose(K, groundTruth))

        K = gym_socks.kernel.metrics.rbf_kernel(Y, Y)
        cls.assertTrue(np.allclose(K, groundTruth))

        K = gym_socks.kernel.metrics.rbf_kernel(Y, distance_fn=euclidean_distances)
        cls.assertTrue(np.allclose(K, groundTruth))

        K = gym_socks.kernel.metrics.rbf_kernel(Y, Y, distance_fn=euclidean_distances)
        cls.assertTrue(np.allclose(K, groundTruth))

    def test_rbf_kernel_different_size(cls):
        """Test that RBF kernel computes correctly with different-sized inputs."""

        X = np.arange(4).reshape((2, 2))
        Y = np.arange(6).reshape((3, 2))

        groundTruth = np.array(
            [
                [
                    [1.0, 0.93941306, 0.77880078],
                    [0.93941306, 1.0, 0.93941306],
                    [0.77880078, 0.93941306, 1.0],
                ],
            ],
            dtype=np.float32,
        )

        K = gym_socks.kernel.metrics.rbf_kernel(Y)
        cls.assertTrue(np.allclose(K, groundTruth))

        K = gym_socks.kernel.metrics.rbf_kernel(Y, Y)
        cls.assertTrue(np.allclose(K, groundTruth))

        groundTruth = np.array(
            [[1.0, 0.93941306, 0.77880078], [0.93941306, 1.0, 0.93941306]],
            dtype=np.float32,
        )

        K = gym_socks.kernel.metrics.rbf_kernel(X, Y)
        cls.assertTrue(np.allclose(K, groundTruth))

        K = gym_socks.kernel.metrics.rbf_kernel(X, Y, distance_fn=euclidean_distances)
        cls.assertTrue(np.allclose(K, groundTruth))

    def test_close_to_sklearn(cls):
        """Assert that result is close to sklearn."""

        Y = np.arange(4).reshape((2, 2))

        D = gym_socks.kernel.metrics.euclidean_distance(Y, squared=True)

        groundTruth = rbf_kernel(Y, gamma=1 / (2 * np.median(D) ** 2))

        K = gym_socks.kernel.metrics.rbf_kernel(Y)
        cls.assertTrue(np.allclose(K, groundTruth))

        K = gym_socks.kernel.metrics.rbf_kernel(Y, Y)
        cls.assertTrue(np.allclose(K, groundTruth))


class TestRegularizedInverse(unittest.TestCase):
    def test_regularized_inverse(cls):
        """Test that regularized inverse computes correctly."""

        Y = np.arange(4).reshape((2, 2))

        groundTruth = np.array(
            [[0.66676608, -0.0081415], [-0.0081415, 0.66676608]],
            dtype=np.float32,
        )

        W = gym_socks.kernel.metrics.regularized_inverse(Y)
        cls.assertTrue(np.allclose(W, groundTruth))

    def test_sklearn_kernels(cls):
        """Sklearn kernels should work with regularized_inverse."""

        Y = np.arange(4).reshape((2, 2))

        for kernel_fn in sklearn_kernel_list:
            with cls.subTest(msg=f"Testing with {type(kernel_fn)}."):

                try:

                    W = gym_socks.kernel.metrics.regularized_inverse(
                        Y, Y, kernel_fn=kernel_fn
                    )

                except Exception as e:
                    cls.fail(f"Kernel {type(kernel_fn)} raised an exception.")
